/* *****************************************************************
 *
 *  $Id: f0m_server.c,v 1.4 2002/09/27 13:31:51 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/27 14:23:39 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */


#include "config.h"
#include "glob.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <netdb.h>
#include <getopt.h>
#include <unistd.h>
#include <assert.h>
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#include <pthread.h>
#include <semaphore.h>

#include "sock_ntop.h"

#include "f0m_tkin.h"
#include "f0m_rw_to_file.h"
#include "f0m_lib.h"




/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

#define NO_HALT	0
#define HALT	1

#define LISTENQ 10

#define VERBOSE(exp) if (options.verbose) msg_pid_write exp


typedef struct options_s options_t;
struct options_s {
    int debug;					/* 0 */
    int verbose;				/* 0 */
    char *service;				/* F0M_SERV_SERVICE_NUMBER */
    char *xdr;					/* NULL */
};

/* Default options values */
options_t options = {0, 0, F0M_SERVER_SERVICE_NUMBER, NULL};


typedef struct client_data {
    int fd;
    pthread_t thid;
    const char *hostaddr;
} sock_data_t;


#define TERM_QUIT   0
#define TERM_HALT	1


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    void f0m__usage ()
 *
 * Description: Print the f0m__usage string and exits
 *
 * Returns:     Never returns!!!
 */
PRIVATE void
f0m__usage (const char *progname)
{
    printf ("\n\
f0m__Usage: %s [-h|-?] [-d] [-v] [-k] [-s <service/port#>] -x [jml|xml]\n\
\n\
Option: -h | -?             Print this help message.\n\
        -d                  Print debug information.\n\
        -v                  Be verbose.\n\
        -k                  Keep server alive.\n\
        -s <service/port#>  Service name / port number.\n\
        -x [jml|xml]        XDR protocol used.\n\
\n\
", progname);
    exit (1);
}



/* -----------------------------------------------------------------
 * Function:    int f0m__check_options ()
 *
 * Description: Check and validate the command line options
 *
 * Returns:     
 */
PRIVATE int
f0m__check_options (int argc, char *argv[])
{

/* OPTION:	s	The general service name / service number */
#define THE_OPITIONS "h?dvks:x:"

    int ch;

    while ((ch = getopt (argc, argv, THE_OPITIONS)) != EOF) {
/* 		assert (ch != '?'); */
        switch (ch) {
        case 'h':
        case '?':
            f0m__usage (argv[0]);
            break;
        case 'd':
            options.debug = 1;
            break;
        case 'v':
            options.verbose = 1;
            break;
        case 's':
            options.service = optarg;
            break;
        case 'x':
            options.xdr = optarg;
			setenv ("FIDDLE_XDR", optarg, 1);
            break;
        }
    }
    return optind;
}


/* -----------------------------------------------------------------
 * Function:    int f0m__update_client_counter ()
 *
 * Description: Updates the counter associated with the number of
 *				active clients
 *
 * Returns:     The number of active clients
 */
PRIVATE int
f0m__update_client_counter (int val)
{
    static int active = 0;
    static pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;
    int local_active;

    pthread_mutex_lock (&m);
    active += val;
    local_active = active;
    pthread_mutex_unlock (&m);
    return local_active;
}



/* -----------------------------------------------------------------
 * Function:    void f0m__terminate ()
 *
 * Description: Terminate Fiddle
 *
 * Returns:     []
 */
PRIVATE void
f0m__terminate (int halt)
{
	/* Update (decrement) the number of clients */
	if (halt  ||  f0m__update_client_counter (-1) == 0) {
        /* No more clients active. Terminate the server */
        f0m_terminate ();
        exit (0);
    }
}



/* -----------------------------------------------------------------
 * Function:    void f0m__sig_handler ()
 *
 * Description: Signal handler
 *
 * Returns:     
 */
PRIVATE void
f0m__sig_handler (int signum)
{
    VERBOSE (("Signal %d received!  Exiting...\n", signum));
    f0m__terminate (HALT);
}



/* -----------------------------------------------------------------
 * Function:    int f0m__tcp_listen ()
 *
 * Description: Create a TCP socket, bind the server's port, and allow 
 *				incomming connection requests to be accepted
 *				(adapted from "Unix Networking Programming", Vol. I, 
 *				 by W. Richard Stevens)
 *
 * Returns:     
 */
PRIVATE int
f0m__tcp_listen (const char *host, const char *serv, socklen_t * addrlenp)
{
    const int on = 1;
    int listenfd, n;
    struct addrinfo hints, *res, *ressave;

    memset (&hints, 0, sizeof (hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = PF_INET;
    hints.ai_socktype = SOCK_STREAM;
    n = getaddrinfo (host, serv, &hints, &res);
    assert (n == 0);

    for (ressave = res; res != NULL; res = res->ai_next) {
        listenfd = socket (res->ai_family,
                           res->ai_socktype, res->ai_protocol);
        if (listenfd > 0) {
            setsockopt (listenfd, SOL_SOCKET, SO_REUSEADDR,
                        &on, sizeof (on));
            if (bind (listenfd, res->ai_addr, res->ai_addrlen) == 0)
                break;          /* Success!!! */
            close (listenfd);
        }
    }
    assert (res != NULL);
    listen (listenfd, LISTENQ);
    if (addrlenp)
        *addrlenp = res->ai_addrlen;
    freeaddrinfo (ressave);
    return listenfd;
}


typedef struct tkin_and_fd_s tkin_and_fd_t;
struct tkin_and_fd_s {
	tkin_t		*tkin;
	int			 fd;
	pthread_t	 thid;
};

/* -----------------------------------------------------------------
 * Function:    void service_thread ()
 *
 * Description: Main loop of the thread created to attend requests
 *				from one client
 *
 * Returns:     Never returns!!!  BUG!
 */
PRIVATE void*
f0m__execute_requested_service (void *_tk)
{
	int err;
	tkout_t *tkout;
	tkin_and_fd_t *tk = (tkin_and_fd_t *)_tk;
	
	assert (tk != NULL);
	assert (tk->tkin != NULL);
	assert (tk->fd >= 0);

	/* Request the service */
	tkout = f0m_tkin3 (tk->tkin->tid, tk->tkin, 1);
	assert (tkout != NULL);

	/* Send the reply */
	err = f0m_write_tkout (tk->fd, tkout);
	assert (err >= 0);

	/* Destroy the output token */
	f0s_tkout_delete (tkout);
	free (tk);

	
	return NULL;
}



/* -----------------------------------------------------------------
 * Function:    void service_thread ()
 *
 * Description: Main loop of the thread created to attend requests
 *				from one client
 *
 * Returns:     Never returns!!!  BUG!
 */
PRIVATE void *
f0m__service_thread (void *arg)
{
    int err;

	/* Check pre-conditions */
    sock_data_t *sd = (sock_data_t *) arg;
    assert (sd->fd >= 0);

	/* Update (increment) the number of clients */
    f0m__update_client_counter (1);

	/* Repeat forever */
    for (;;) {
		tkin_and_fd_t *tk = malloc (sizeof (*tk));
		assert (tk != NULL);
		tk->fd = sd->fd;
		
        /* Get a request ('tkin_t' structure) */
		VERBOSE (("%s::%s():: Looking in %d\n\n",
					   __FILE__, __FUNCTION__, sd->fd));
        tk->tkin = f0m_read_tkin (sd->fd);
		VERBOSE (("%s::%s():: Got request from %d\n\n",
					   __FILE__, __FUNCTION__, sd->fd));
        if (tk->tkin == NULL)		/* EOF = terminate thread */
            break;

		pthread_create (&(tk->thid), NULL, f0m__execute_requested_service, tk);
		pthread_detach (tk->thid);
    }

	/* Cleanup and terminate thread (all, if last thread)*/
	free (sd);
    f0m__terminate (NO_HALT);
    return NULL;
}




/* -----------------------------------------------------------------
 * Function:    void f0m__setup_server ();
 *
 * Description: Setup a server listening in <service>
 *				Launches a new thread to process requests from each client
 *
 * Returns:     Never returns!!!
 */
PRIVATE void
f0m__setup_server (const char *service)
{
    int listenfd;
    socklen_t addrlen, len;
    struct sockaddr *cliaddr;

    listenfd = f0m__tcp_listen (NULL, service, &addrlen);
    assert (listenfd >= 0);
    assert (addrlen > 0);
    cliaddr = alloca (addrlen);

    for (;;) {
		/* Wait for a client connection */
        sock_data_t *sd = alloca (sizeof (*sd));
        len = addrlen;
        sd->fd = accept (listenfd, cliaddr, &len);
		assert (sd->fd >= 0);

		/* Get client info */
        sd->hostaddr = sock_ntop (cliaddr, len);
        VERBOSE (("%s::f0m__setup_server()::Connection from %s\n",
				  __FILE__, sd->hostaddr));

		/* Start handler thread */
        pthread_create (&(sd->thid), NULL, f0m__service_thread, sd);
    }
}



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    main()
 *
 * Description: The main function
 *
 * Returns:     != 0 when terminating with error
 */
PUBLIC int
main (int argc, char *argv[])
{
	/* Check options */
    int optind = f0m__check_options (argc, argv);
    VERBOSE (("%s::SERVICE NAME/PORT# = %s\n", __FILE__, options.service));
    if (optind != argc)
        f0m__usage (argv[0]);   /* Calls 'exit (1)' */

	/* Activate signal handlers */
    if (signal (SIGINT, f0m__sig_handler) == SIG_ERR)
        assert (0);

	/* Initialize Fiddle (0m) */
    f0m_initialize ();

	/* Accept client connections  */
    f0m__setup_server (options.service);

	/* Will never be reached, just to avoid compiler warinings */
    return 0;
}
