/* *****************************************************************
 *
 *  $Id: f0s_dbase.c,v 1.4 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/25 15:50:26 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Local database with information about the debuggers
 *				 running in the local machine
 *
 * -----------------------------------------------------------------
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>			/* for mkfifo() */
#include <sys/stat.h>			/* for mkfifo() */
#include <assert.h>
#include "glob.h"
#include "msg.h"
#include "chain.h"
#include "f0s_lib.h"
#include "f0s_dbg_io.h"
#include "f0s_getdelim.h"
#include "f0s_dbase.h"
#include "config.h"


/* Define one and only one of the two below */
/* #define F0S_USE_PIPE */
#define F0S_USE_FIFO


#define TEMP_DIR_NAME "/tmp"

typedef struct dbginfo_s dbginfo_t;

struct dbginfo_s {
	int tid;					/* The debugger's ID */
	int dbg_pid;				/* The debugger's PID */
	int tp_pid;					/* The PID of the target process */
	int attached;				/* True if CLD was attached to process */
	FILE *wfd;					/* The file descriptor to write commands
								 * to the CLD */
	FILE *rfd;					/* The file descriptor to read replies
								 * from the CLD */
	char *wfile;				/* The name of the fifos to send and */
	char *rfile;				/* receive data to/from the CLD, empty
								   strings if fifos are not being used */
};



PRIVATE chain_t	*db_tid = NULL;



PRIVATE char *gdb_initialization[] = {
	"set print null-stop",
	"set print pretty",
	"set confirm off",
	NULL,
};


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


#define CHECK_INITIALIZED								\
	if (! db_tid) 										\
		db_tid = chain_create ()



/* -----------------------------------------------------------------
 * Function:    void setup_new_cld ()
 *
 * Description: Initialize the new CLD
 *
 * Returns:     []
 */
PRIVATE void f0s__setup_new_cld (int tid)
{
#define MAX_BUF 256
	int i;
	for (i = 0;  gdb_initialization[i] != NULL;  i++) {
		char *rep = f0s_send_receive (tid, gdb_initialization[i]);
		if (rep != NULL) free (rep);
	}
}




/* -----------------------------------------------------------------
 * Function:    int rel = f0s_dbase_cmp ()
 *
 * Description: Compare two 'DbgInfo' variables.
 *
 * Returns:		-1		If (*left < *rigth)
 *				 0		If (*left == *rigth)
 *				 1		If (*left > *rigth)
 */
#define CMP_TID			0
#define CMP_DBG_PID		1
#define CMP_TP_PID		2
#define CMP_ATTACHED	3

PRIVATE int f0s__dbase_cmp (const void *vdata, const void *vwhat, void *vargs)
{
	int v1, v2;
	int args = (int)vargs;
	const dbginfo_t *l = (const dbginfo_t *)vdata;
	const int *r = (int *)vwhat;
	switch (args) {
	case CMP_TID:				/* Compare debugger IDs */
		v1 = l->tid;
		break;
	case CMP_DBG_PID:			/* Compare debugger PID */
		v1 = l->dbg_pid;
		break;
	case CMP_TP_PID:			/* Compare target process PID */
		v1 = l->tp_pid;
		break;
	case CMP_ATTACHED:			/* Compare attached status */
		v1 = l->attached;
		break;
	default:
		msg_pid_fatal ("f0s_dbase_cmp(): Invalid item to cmpare '%d'\n", 
					   args);
	}
	v2 = *r;
	if (v1 < v2)
		return -1;
	if (v1 > v2)
		return 1;
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    DbgInfo *info = f0s_dbase_get_dbginfo ()
 *
 * Description: Get the 'DbgInfo' structure associated with 'tid'
 *
 * Returns:     The 'DbgInfo' structure
 */
PRIVATE dbginfo_t *f0s__dbase_get_dbginfo (int tid)
{
	chain_node_t *link;
	CHECK_INITIALIZED;
	link =  chain_find (db_tid, &tid, f0s__dbase_cmp,
						(void *)CMP_TID);
	if (link == NULL) return NULL; /* Not found */
	return (dbginfo_t *)chain_node_data(link);	/* Found */
}


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_new ()
 *
 * Description: Create a new entry in the database for a new CLD
 *
 * Returns:     The debugger's identifier
 */
PRIVATE int f0s__dbase_new (int dbg_pid, int tp_pid, int rfd, int wfd,
							const char *rfile, const char *wfile)
{
	PRIVATE int counter = 0;
	chain_node_t *link;
	dbginfo_t *info;
	
	CHECK_INITIALIZED;
	info = malloc (sizeof (*info));
	if (info == NULL) return -1; /* Memory error */

	/* Fill the structure.
	 * The 'tid' must be filled with exclusive access */
	info->dbg_pid = dbg_pid;
	info->tp_pid = tp_pid;
	info->attached = 0;
	if ((info->rfd = fdopen (rfd, "r")) == NULL) return -1;
	if ((info->wfd = fdopen (wfd, "w")) == NULL) return -1;
	if (setvbuf(info->wfd, NULL, _IONBF, 0) < 0)
		return -1;
	info->rfile = strdup (rfile);
	assert (info->rfile != NULL);
	info->wfile = strdup (wfile);
	assert (info->rfile != NULL);
	/* Request exclusive access */
	info->tid = ++counter;
	link = chain_enqueue (db_tid, info);
	/* Freed exclusive access */
	if (link == NULL) {
		if (info->rfd != NULL) fclose (info->rfd);
		if (info->wfd != NULL) fclose (info->wfd);
		free (info);
		return -1;
	}
	return info->tid;
}




/* -----------------------------------------------------------------
 * Function:    int tid = f0s__dbase_free ()
 *
 * Description: Free all data in a node of the database
 *
 * Returns:     Pointer to dbase entry
 */
PRIVATE int f0s__dbase_free (chain_node_t *node)
{
	int status;
	tkout_t *tkout;
	dbginfo_t *info;

	assert (node !=NULL);
	info = chain_node_data (node);
	assert (info != NULL);

	tkout = info->attached ? f0s_detach (info->tid) : f0s_kill (info->tid);
	f0s_tkout_delete (tkout);

	/* Free file descriptors */
	fclose (info->wfd);
	fclose (info->rfd);

	/* Free fifo entries in TEMP dir */
	status = unlink (info->wfile);
/* 	assert (status == 0); */
	free (info->wfile);
	status = unlink (info->rfile);
/* 	assert (status == 0); */
	free (info->rfile);

	/* Remove entry from database */
	chain_remove (db_tid, node);
}

/* -----------------------------------------------------------------
 * Function:    int tid = f0s__create_communication_channels ()
 *
 * Description: Create communication channels for data exchange
 *				between the client and a CLD
 *
 * Returns:     0 if OK, -1 on error
 */
PRIVATE int f0s__create_communication_channels (int fd_cmd[], int fd_rep[],
												char *cmd_file,
												char *rep_channel)
{
#if defined F0S_USE_PIPE
	if (pipe (fd_cmd) < 0) return -1;
	if (pipe (fd_rep) < 0) return -1;
	strcpy (cmd_file, "");
	strcpy (rep_channel, "");

#elif defined F0S_USE_FIFO
	char *status_c;
	int status_i;
	char *temp_dir_name = getenv ("TEMP");
	if (temp_dir_name == NULL)
		temp_dir_name = getenv ("TMP");
	if (temp_dir_name == NULL)
		temp_dir_name= TEMP_DIR_NAME;

	/* Create fifo to send commands to the CLD */
	strcpy (cmd_file, temp_dir_name);
	strcat (cmd_file, "/cmd-XXXXXX");
	status_c = mktemp (cmd_file);
	assert (status_c);
	status_i = mkfifo (cmd_file, 0600);
	assert (status_i == 0);

	/* Create fifo to receive replies from CLD */
	strcpy (rep_channel, temp_dir_name);
	strcat (rep_channel, "/rep-XXXXXX");
	status_c = mktemp (rep_channel);
	assert (status_c);
	status_i = mkfifo (rep_channel, 0600);
	assert (status_i == 0);

#else
#error Please define the communication mechanism to use: F0S_USE_PIPE or F0S_USE_FIFO!
#endif

	return 0;					/* OK */
}


/* -----------------------------------------------------------------
 * Function:    int tid = f0s__open_communication_channels ()
 *
 * Description: Open the communication channels for data exchange
 *				between the client and a CLD
 *
 * Returns:     0 if OK, -1 on error
 */
PRIVATE int f0s__open_communication_channels (int fiddle_side,
											  int fd_cmd[], int fd_rep[],
											  const char *cmd_file,
											  const char *rep_channel)
{
#if defined F0S_USE_PIPE
	if (fiddle_side) {		/* is the father */
		/* Close the pipe entries not used */
		close (fd_cmd[0]);
		close (fd_rep[1]);
	}
	else {						/* is the son */
		/* Redirect IO to pipe */
		dup2 (fd_cmd[0], 0);
		dup2 (fd_rep[1], 1);
		dup2 (fd_rep[1], 2);
		close (fd_cmd[0]);
		close (fd_cmd[1]);
		close (fd_rep[0]);
		close (fd_rep[1]);
	}
#elif defined F0S_USE_FIFO
	if (fiddle_side) {		/* is the father */
		/* Open the fifos */
#ifdef EBUG
		msg_pid_write ("FIDDLE_SIDE: Opening '%s' for writing\n", cmd_file);
#endif
		fd_cmd[1] = open (cmd_file, O_WRONLY);
		assert (fd_cmd[1] != -1);
#ifdef EBUG
		msg_pid_write ("FIDDLE_SIDE: Opening '%s' for reading\n", rep_channel);
#endif
		fd_rep[0] = open (rep_channel, O_RDONLY);
		assert (fd_rep[0] != -1);
	}
	else {						/* is the son */
		/* Open the fifos */
/* 		msg_pid_write ("GDB_SIDE: Opening '%s' for reading\n", cmd_file); */
		fd_cmd[0] = open (cmd_file, O_RDONLY);
		assert (fd_cmd[0] != -1);
/* 		msg_pid_write ("GDB_SIDE: Opening '%s' for writing\n", rep_channel); */
		fd_rep[1] = open (rep_channel, O_WRONLY);
		assert (fd_rep[1] != -1);
		/* Redirect IO to pipe */
		dup2 (fd_cmd[0], 0);
		dup2 (fd_rep[1], 1);
		dup2 (fd_rep[1], 2);
		close (fd_cmd[0]);
		close (fd_rep[1]);
	}
#else
#error Please define the communication mechanism to use: F0S_USE_PIPE or F0S_USE_FIFO!
#endif
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    int tid = f0s__start_new_cld ()
 *
 * Description: Start a new CLD (gdb)
 *
 * Returns:     The 'tid' of the newly created CLD
 */
PRIVATE int f0s__start_new_cld (void)
{
#ifndef PATH_MAX
#define PATH_MAX 128
#endif
	int pid;
	int tid;
	int fd_cmd[2];			/* Send commands to CLD */
	int fd_rep[2];			/* Get replies from CLD */
	char cmd_channel[PATH_MAX];	/* Send commands to CLD*/
	char rep_channel[PATH_MAX];	/* Get replies from CLD */
	char *prompt;
	int status;

	/* Create communication channels */
	status = f0s__create_communication_channels (fd_cmd, fd_rep,
												 cmd_channel, rep_channel);
	assert (status == 0);

	/* Create new process. The son will redired IO to communication
	   channels and exec the debugger. The parent will register the
	   communication channels in the database */
	pid = fork ();
	switch (pid) {
	case -1:				/* Error */
		msg_pid_fatal ("Error on f0s_dbase_find_unused()::fork()");
	case 0:					/* Son */
		status = f0s__open_communication_channels (pid, fd_cmd, fd_rep,
												   cmd_channel, rep_channel);
		assert (status == 0);
		fcntl(1, F_SETFL, O_APPEND);
		if (setvbuf(stdout, NULL, _IONBF, 0) < 0)
			return -1;
/*  		execlp ("mygdb", "./gdb", NULL); */
 		execlp ("gdb", "gdb", "--fullname", "-q", NULL);
		msg_pid_fatal ("Error on f0s_dbase_find_unused_unused()::execlp()");
	default:				/* Father */
		status= f0s__open_communication_channels (pid, fd_cmd, fd_rep,
												  cmd_channel, rep_channel);
		assert (status == 0);
	
		/* Create a new entry in the database */
		tid = f0s__dbase_new (pid, -1, fd_rep[0], fd_cmd[1],
							  rep_channel, cmd_channel);

		/* Wait for the CLD's prompt */
		prompt = get_delim (f0s_dbase_get_rfd (tid), "(gdb) ");
		free (prompt);
	
		/* Set up some parameters */
		f0s__setup_new_cld (tid);
		break;
	}
	return tid;
}



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_add_cld ()
 *
 * Description: Add a cld + process to the database
 *
 * Returns:     Tid of new process
 */
PUBLIC int f0s_dbase_add_cld (int dbg_pid, int tp_pid,
							  const char *cmd_channel, const char *rep_channel)
{
	int status;
	int tid;
	char *prompt;
	int fd_cmd[2];
	int fd_rep[2];
#define FIDDLE_SIDE 1
	
#if defined F0S_USE_PIPE
	assert (0);					/* Not implemented */
#elif defined F0S_USE_FIFO
	status= f0s__open_communication_channels (FIDDLE_SIDE,
											  fd_cmd, fd_rep,
											  cmd_channel, rep_channel);
#else
#error Please define the communication mechanism to use: F0S_USE_PIPE or F0S_USE_FIFO!
#endif
	assert (status == 0);
		
	/* Create a new entry in the database */
	tid = f0s__dbase_new (dbg_pid, tp_pid, fd_rep[0], fd_cmd[1],
						  rep_channel, cmd_channel);
	
	/* Wait for the CLD's prompt */
	prompt = get_delim (f0s_dbase_get_rfd (tid), "(gdb) ");
	free (prompt);

	/* Set up some parameters */
	f0s__setup_new_cld (tid);
	return tid;
}


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_cleanup ()
 *
 * Description: Remove all entries in the database
 *
 * Returns:     Status
 */
PUBLIC int f0s_dbase_cleanup (void)
{
	/* Remove all dbuggers from database */
	while (f0s_dbase_delete (-1) == 0)
		;
}


/* -----------------------------------------------------------------
 * Function:    int status = f0s_dbase_delete ()
 *
 * Description: Remove a debugger from the database
 *
 * Returns:     The status of the operation
 */
PUBLIC int f0s_dbase_delete (int tid)
{
	chain_node_t *link = NULL;
	CHECK_INITIALIZED;
	if (tid == -1)
		link = chain_first (db_tid);
	else
		link =  chain_find (db_tid, (const void *)&tid, f0s__dbase_cmp,
							(void *)CMP_TID);
	if (link != NULL)
		f0s__dbase_free (link);

	/* Return 0 if OK, -1 on error */
	return ((link != NULL) - 1);
}




/* ################################################################# */



/* -----------------------------------------------------------------
 * Function:    int f0s_dbase_list ()
 *
 * Description: List processes currently under debugging
 *
 * Returns:     Chain with data
 */
PUBLIC chain_t *f0s_dbase_list (void)
{
	chain_node_t *node;
	dbginfo_t *info;
	rtids_t *tids;
	chain_t *ch;

	CHECK_INITIALIZED;
	ch = chain_create ();
	assert (ch != NULL);
	for (node = chain_first (db_tid);
		 node != NULL;  node = chain_next (db_tid, node)) {
		tids = malloc (sizeof (*tids));
		assert (tids != NULL);
		info = chain_node_data (node);
		assert (info->attached != -1);
		tids->gtid = info->tid;
		tids->ltid = -1;
		tids->machine = NULL;
		tids->tp_pid = info->tp_pid;
		tids->dbg_pid = info->dbg_pid;
		tids->attached = info->attached;
		chain_enqueue (ch, tids);
	}
	return ch;
}


/* -----------------------------------------------------------------
 * Function:    int f0s_dbase_is_unused ()
 *
 * Description: Check if 'tid' corespond to a unused debugger
 *
 * Returns:     
 */
PUBLIC int f0s_dbase_is_unused (int tid)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL  ||  info->tp_pid != -1)
		/* Non-existatent or being used */
		return 0;
	return 1;					/* Unused */
}


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_unused ()
 *
 * Description: Find an unused debugger (with 'info.attached = -1')
 *
 * Returns:     
 */
PUBLIC int f0s_dbase_find_unused (void)
{
	/* Start a new CLD! */
	int tid = f0s__start_new_cld ();
	assert (tid >= 0);

	/* Set up some parameters */
	f0s__setup_new_cld (tid);

	return tid;
}


/* ################################################################# */


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_by_dbg_pid ()
 *
 * Description: Find the debugger's ID given its PID
 *
 * Returns:     
 */
PUBLIC int f0s_dbase_find_by_dbg_pid (int dbg_pid)
{
	chain_node_t *link;
	dbginfo_t *info;
	CHECK_INITIALIZED;
	link =  chain_find (db_tid, &dbg_pid, f0s__dbase_cmp,
						(void *)CMP_DBG_PID);
	if (link == NULL) return -1; /* Not found */
	/* Found */
	info = (dbginfo_t *)chain_node_data(link);
	return info->tid;
}


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_by_tp_pid ()
 *
 * Description: Find the debugger's ID given the target process's PID
 *
 * Returns:     
 */
PUBLIC int f0s_dbase_find_by_tp_pid (int tp_pid)
{
	chain_node_t *link;
	dbginfo_t *info;
	CHECK_INITIALIZED;
	link = chain_find (db_tid, &tp_pid, f0s__dbase_cmp,
					   (void *)CMP_TP_PID);
	if (link == NULL) return -1; /* Not found */
	/* Found */
	info = (dbginfo_t *)chain_node_data(link);
	return info->tid;
}


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_by_attached ()
 *
 * Description: Find the debugger's ID given the target process's PID
 *
 * Returns:     
 */
PUBLIC int f0s_dbase_find_by_attached (int attached)
{
	chain_node_t *link;
	dbginfo_t *info;
	CHECK_INITIALIZED;
	link = chain_find (db_tid, &attached, f0s__dbase_cmp,
					   (void *)CMP_ATTACHED);
	if (link == NULL) return -1; /* Not found */
	/* Found */
	info = (dbginfo_t *)chain_node_data(link);
	return info->tid;
}


/* ################################################################# */



/* -----------------------------------------------------------------
 * Function:    int tp_pid = f0s_dbase_get_tp_pid ()
 *
 * Description: Get a debugger's target process PID
 *
 * Returns:     The debugger's target process PID
 */
PUBLIC int f0s_dbase_get_tp_pid (int tid)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL) return -1; /* Not found */
	return info->tp_pid;
}


/* -----------------------------------------------------------------
 * Function:    int status = f0s_dbase_set_tp_pid ()
 *
 * Description: Set a debugger's target process PID
 *
 * Returns:     The status of the operation
 */
PUBLIC int f0s_dbase_set_tp_pid (int tid, int tp_pid)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL) return -1; /* Not found */
	info->tp_pid = tp_pid;
	return 0;					/* OK */
}


/* -----------------------------------------------------------------
 * Function:    int status = f0s_dbase_set_attached ()
 *
 * Description: Set a debugger's target process PID
 *
 * Returns:     The status of the operation
 */
PUBLIC int f0s_dbase_set_attached (int tid, int attached)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL) return -1; /* Not found */
	info->attached = attached;
	return 0;					/* OK */
}


/* ################################################################# */


/* -----------------------------------------------------------------
 * Function:    FILE *rfd = f0s_dbase_get_rfd ()
 *
 * Description: Get the channel that allows to receive data from the
 *				    debugger
 *
 * Returns:     The FILE descriptor
 */
PUBLIC FILE *f0s_dbase_get_rfd (int tid)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL) return NULL; /* Not found */
	return info->rfd;
}


/* -----------------------------------------------------------------
 * Function:    FILE *wfd = f0s_dbase_get_wfd ()
 *
 * Description: Get the channel that allows to send data to the
 *				    debugger
 *
 * Returns:     The FILE descriptor
 */
PUBLIC FILE *f0s_dbase_get_wfd (int tid)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL) return NULL; /* Not found */
	return info->wfd;
}


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_get_wfile ()
 *
 * Description: Get the name of the fifo usedto communicate with the CLD
 *
 * Returns:     A pointer to the name
 */
PUBLIC char *f0s_dbase_get_wfile (int tid)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL) return NULL; /* Not found */
	return info->wfile;
}


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_get_rfile ()
 *
 * Description: Get the name of the fifo usedto communicate with the CLD
 *
 * Returns:     A pointer to the name
 */
PUBLIC char *f0s_dbase_get_rfile (int tid)
{
	dbginfo_t *info = f0s__dbase_get_dbginfo (tid);
	if (info == NULL) return NULL; /* Not found */
	return info->rfile;
}
