/* *****************************************************************
 *
 *  $Id: f1m_queues.c,v 1.3 2002/08/14 01:30:25 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/26 23:34:03 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implementation for the enqueue requests and dequeue
 *				 replies services
 *
 *		f0m_queues_init ()		-- Initialization
 *		f0m_queues_terminate ()	-- Termination
 *		f0m_enqueue_request ()	-- Enqueue a request
 *		f0m_dequeue_request ()	-- Dequeue a request
 *		f0m_enqueue_reply ()		-- Enqueue a reply
 *		f0m_dequeue_reply ()		-- Dequeue a reply
 *
 * -----------------------------------------------------------------
 */

#include <stdlib.h>
#include <assert.h>
#include <pthread.h>

#include "chain.h"

#include "f1m_servdata.h"
#include "f1m_queues.h"

#include "config.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */


PRIVATE chain_t *req_q = NULL;
PRIVATE chain_t *rep_q = NULL;


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    unsigned long reqid = generate_new_request_id (void)
 *
 * Description: Generate a new request ID
 *
 * Returns:     The new request ID
 */
static unsigned long generate_new_request_id (void)
{
	static pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
	static unsigned long the_id = 0;
	unsigned long reqid ;
	
	pthread_mutex_lock (&mut);
	reqid = ++the_id;
	pthread_mutex_unlock (&mut);
	return reqid;
}


/* -----------------------------------------------------------------
 * Function:    int cmp_id_f (void *data, void *args)
 *
 * Description: Test 'if (((reply_t *)data)->id == (int) args)'
 *
 * Returns:     1 if condition evaluates to TRUE
 *				0 otherwise
 */
static int cmp_id_f (void *data, void *args)
{
	assert (data != NULL);
	if (((reply_t *)data)->id == (int) args)
		return 1;				/* TRUE */
	return 0;					/* FALSE */
}


/* -----------------------------------------------------------------
 * Function:    int cmp_servid_f (void *data, void *args)
 *
 * Description: Test 'if (((reply_t *)data)->servid == (int) args)'
 *
 * Returns:     1 if condition evaluates to TRUE
 *				0 otherwise
 */
static int cmp_servid_f (void *data, void *args)
{
	assert (data != NULL);
	if (((reply_t *)data)->servid == (int) args)
		return 1;				/* TRUE */
	return 0;					/* FALSE */
}


/* -----------------------------------------------------------------
 * Function:    request_t *req = tkin_2_request (int servid,
 *												 const tkin_t *tk)
 *
 * Description: Generate a request ID and pack it with the 'tkin_t'
 *				token into a 'request_t' structure.
 *
 * Returns:     A pointer to a newly allocated 'request_t' structure
 */
static request_t *tkin_2_request (int servid, const tkin_t *tk)
{
	request_t *req = malloc (sizeof (*req));
	assert (req != NULL);
	req->id = generate_new_request_id ();
	req->servid = servid;
	req->tk = tk;
	return req;
}


/* -----------------------------------------------------------------
 * Function:    reply_t *rep = tkout_2_reply (int servid, int id, 
 *											  const tkout_t *tk)
 *
 * Description: Pack the request ID, task ID and the reply token into
 *				a 'reply_t' structure.
 *
 * Returns:     A pointer to a newly allocated 'reply_t' structure
 */
static reply_t *tkout_2_reply (int servid, int id, const tkout_t *tk)
{
	reply_t *rep = malloc (sizeof (*rep));
	assert (rep != NULL);
	rep->id = id;
	rep->servid = servid;
	rep->tk = tk;
	return rep;
}




/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    int f1m_queues_init (void)
 *
 * Description: Initialize queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
int f1m_queues_init (void)
{
	assert (req_q == NULL);
	assert (rep_q == NULL);
	req_q = chain_create ();
	rep_q = chain_create ();
	assert (req_q != NULL);
	assert (rep_q != NULL);
	return 0;
}

/* -----------------------------------------------------------------
 * Function:    int f1m_queues_terminate (void)
 *
 * Description: Terminate queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
int f1m_queues_terminate (void)
{
	assert (req_q != NULL);
	assert (rep_q != NULL);
	chain_destroy (req_q);
	chain_destroy (rep_q);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    int f1m_enqueue_request (int gtid, const tkin_t *req)
 *
 * Description: 'gtid' The (symbolic) id of the target process
 *				'req'	 Token with the request code and arguments
 *
 * Returns:     The request ID
 */
int f1m_enqueue_request (int gtid, const tkin_t *tk)
{
	PRIVATE serv_data_tid_t no_serv = {-1, -1, -1, {-1, -1}};
	request_t *req;
	void *err;
	serv_data_tid_t *data;

	/* Find the 'servid' associated with 'gtid' */
	if (gtid == -1)
		data = &no_serv;
	else
		data = f1m_tp_gtid_2_servdata (gtid);
	if (data == NULL)
		return -1;				/* Invalid TID */

	/* Enqueue request */
	req = tkin_2_request (data->servid, tk);
	assert (req_q != NULL);
	err = chain_enqueue (req_q, req);
	assert (err != NULL);
	return req->id;
}


/* -----------------------------------------------------------------
 * Function:    request_t *req f1m_dequeue_request (int servid)
 *
 * Description: 'servid'	The (symbolic) id of the target process
 *
 * Returns:     A structure 'request_t'
 */
request_t *f1m_dequeue_request (int servid)
{
	request_t *req;

	assert (req_q != NULL);
	req = (request_t *) chain_dequeue_cond (req_q, cmp_servid_f, (void *)servid);
	return req;
}


/* -----------------------------------------------------------------
 * Function:    int repid = f1m_enqueue_reply (int gtid, int id,
 *											  const tkout_t *rep)
 *
 * Description: 'gtid'	The (symbolic) id of the target process
 *				'rep'	Token with the reply code and arguments
 *
 * Returns:     The status
 */
int f1m_enqueue_reply (int gtid, int id, const tkout_t *tk)
{
	reply_t *rep;
	void *err;

	rep = tkout_2_reply (gtid, id, tk);
	assert (rep_q != NULL);
	err = chain_enqueue (rep_q, rep);
	assert (err != NULL);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *tk = f1m_dequeue_reply (int reqid)
 *
 * Description: Get the reply to the request with ID 'reqid'
 *
 * Returns:     Pointer to the reply ('tkout_t') structure
 */
tkout_t *f1m_dequeue_reply (int reqid)
{
	reply_t *rep;				/* The reply structure */

	assert (rep_q != NULL);
	rep = (reply_t *) chain_dequeue_cond (rep_q, cmp_id_f, (void *)reqid);
	return (tkout_t *)rep->tk;
}

