/* *****************************************************************
 *
 *  $Id: f1m_servdata.c,v 1.3 2002/08/12 00:29:35 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/04/03 16:59:12 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Mantain a database that maps 'Global Task IDs' into
 *				     'Local Task IDs' and the I/O channels to the
 *				     associated F1 server.
 *
 *		f1m_servdata_initialize ()	-- Initialize the database
 *		f1m_servdata_terminate ()    -- Terminate the database
 *		f1m_gtid_2_servdata ()		-- Get data from database
 *		f1m_add_servdata ()		    -- Add data to the database
 *
 * -----------------------------------------------------------------
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <assert.h>
#include <stdio.h>

#include "chain.h"

#include "f0s_tkout.h"

#include "f0m_tkin.h"
#include "f0m_rw_to_file.h"

#include "f1m_servdata.h"

#include "config.h"



/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

/* Data about processes under debugging */
typedef struct target_process_data_s {
	chain_t *tids;				/* TaskIDs of processes under
								   debugging */
	chain_t *machines;			/* List of machines with processes
								   under debug */
} target_process_data_t;

PRIVATE target_process_data_t tp_data = {NULL, NULL};


/* Data about fiddle clients */
typedef struct fiddle_clients_data_s {
	chain_t *data;				/*  Each node will contain:
									- int  pid
									- char *filename
									- char *hostaddr
								 */
} fiddle_clients_data_t;

PRIVATE fiddle_clients_data_t fc_data = {NULL};



void *f1m__tkout_reader(void *args);
int f1m__tkout_id_cmp( void *data, void *args);

/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    f1m__fc_myfree_data ()
 *
 * Description: Release the memory used by 'obj'
 *
 * Returns:     0
 */
PRIVATE int
f1m__fc_myfree_client_data (void *obj, void *args)
{
	client_data_t *d = (client_data_t *)obj;
	assert (obj != NULL),
	free (d->filename);
	free (d->hostaddr);
	free (d);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:     f1m__fc_find_client_by_pid_and_delete()
 *
 * Description: 
 *
 * Returns:     1 if equal, 0 otherwise
 */
PRIVATE int
f1m__fc_find_client_by_pid_and_delete (void *data, void *arg)
{
	client_data_t *fc = (client_data_t *)data;
	int *pid = (int *)arg;
	if (fc->pid == *pid) {
		f1m__fc_myfree_client_data (fc, NULL);
		return 1;
	}
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    f1m__tp_myfree_tids ()
 *
 * Description: Release the memory used by 'obj' of type 'serv_data_tid_t'
 *
 * Returns:     0
 */
PRIVATE int
f1m__tp_myfree_tids (void *obj, void *args)
{
	free ((serv_data_tid_t *)obj);
	return 0;
}

/* -----------------------------------------------------------------
 * Function:    f1m__tp_myfree_machines ()
 *
 * Description: Release the memory used by 'obj'
 *
 * Returns:     0
 */
PRIVATE int
f1m__tp_myfree_machines (void *obj, void *args)
{
	serv_data_machine_t *d = (serv_data_machine_t *)obj;
	close (d->fd[0]);
	close (d->fd[1]);
	free (d->machine);
	free (d);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    f1m__tp_find_by_sltid ()
 *
 * Description: 
 *
 * Returns:     Pointer to data or NULL if not found
 */
PRIVATE int
f1m__tp_find_by_sltid (const void *data, const void *what, void *arg)
{
	serv_data_tid_t *td = (serv_data_tid_t *)data;
	int *sltid = (int *)what;
	if (td->servid == sltid[0]  &&  td->ltid == sltid[1] )
		return 0;
	return -1;
}


/* -----------------------------------------------------------------
 * Function:    f1m__tp_find_by_gtid ()
 *
 * Description: Check if '(*(int *)what) == ((serv_data_tid_t *)data).gtid'
 *
 * Returns:     Pointer to data or NULL if not found
 */
PRIVATE int
f1m__tp_find_by_gtid (const void *data, const void *what, void *arg)
{
	if ((*(int *)what) == ((serv_data_tid_t *)data)->gtid)
		return 0;
	return -1;
}


/* -----------------------------------------------------------------
 * Function:    f1m__tp_find_by_machine ()
 *
 * Description: Check if '((char *)what) == ((serv_data_machine_t *)data).machine'
 *
 * Returns:     Pointer to data or NULL if not found
 */
PRIVATE int
f1m__tp_find_by_machine (const void *data, const void *what, void *arg)
{
	if (strcmp ((char *)what, ((serv_data_machine_t *)data)->machine) == 0)
		return 0;
	return -1;
}


/* -----------------------------------------------------------------
 * Function:    f1m__tp_find_by_servid ()
 *
 * Description: Check if '((int)what) == ((serv_data_machine_t *)data).servid'
 *
 * Returns:     Pointer to data or NULL if not found
 */
PRIVATE int
f1m__tp_find_by_servid (const void *data, const void *what, void *arg)
{
	if (*((int *)what) == ((serv_data_machine_t *)data)->servid)
		return 0;
	return -1;
}


/* -----------------------------------------------------------------
 * Function:    f1m__tp_apply_tkin ()
 *
 * Description: Apply a service request to a machine.
 *
 * Returns:     The reply token.
 */
PRIVATE void *
f1m__tp_apply_tkin (void *obj, void *args)
{
	serv_data_machine_t *data = (serv_data_machine_t *)obj;
	tkin_t *tkin = (tkin_t *)args;
	tkout_t *tkout;
	serv_data_tid_t *td;
	int err;
	int tid;

	/* Send the request to the server */
	err = f0m_write_tkin (data->fd[1], tkin);
	assert (err > 0);

	/* Get the reply */

	tid = 0;
	tkout = (tkout_t *)chain_dequeue_cond(data->p_reps, f1m__tkout_id_cmp, tid);
	assert (tkout != NULL);

	/* Return the reply token */
	if (tkin->tid != -1) {
		td = f1m_tp_sltid_2_tiddata (data->servid, tkin->tid);
		assert (td != NULL);
		tkout->tid = td->gtid;
	}
	return tkout;
}





/* #################################################################
 *		PUBLIC functions

 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    f1m_servdata_initialize ()
 *
 * Description: Initialize the server information database
 *
 * Returns:     0
 */
PUBLIC int
f1m_servdata_initialize (void)
{
	tp_data.machines = chain_create ();
	assert (tp_data.machines != NULL);
	tp_data.tids = chain_create ();
	assert (tp_data.tids != NULL);
	fc_data.data = chain_create ();
	assert (fc_data.data != NULL);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    f1m_servdata_terminate ()
 *
 * Description: Terminate the server information database
 *
 * Returns:     0
 */
PUBLIC int
f1m_servdata_terminate (void)
{
	chain_node_t *p = chain_prune (tp_data.tids, f1m__tp_myfree_tids, NULL);
	assert (p == NULL);			/* Be sure all nodes were traversed */
	chain_destroy (tp_data.tids);
	p = chain_prune (tp_data.machines, f1m__tp_myfree_machines, NULL);
	assert (p == NULL);			/* Be sure all nodes were traversed */
	chain_destroy (tp_data.machines);
	p = chain_prune (fc_data.data, f1m__fc_myfree_client_data, NULL);
	assert (p == NULL);			/* Be sure all nodes were traversed */
	chain_destroy (fc_data.data);
	return 0;
}



/* -----------------------------------------------------------------
 * Function:    f1m_fc_clients ()
 *
 * Description: Return a pointer to a chain with client data
 *
 * Returns:     Pointer to the chain with data or NULL on error
 */
PUBLIC chain_t *
f1m_fc_clients (void)
{
	return fc_data.data;
}



/* -----------------------------------------------------------------
 * Function:    f1m_fc_add_client_data ()
 *
 * Description: Add data to the 'clients' database
 *
 * Returns:     Pointer to the node with data or NULL on error
 */
PUBLIC int
f1m_fc_add_client_data (int pid, const char *filename, const char *hostaddr)
{
	chain_node_t *n;
	char hostname[128];
	char *pos;
	struct hostent *he;
	client_data_t *d = malloc (sizeof (*d));
	assert (d != NULL);
	if (d == NULL)
		return -1;

	/* 'hostaddr' is like 'hostname/port'
	    Extract only hostname */
	strcpy (hostname, hostaddr);
	pos = strchr (hostname, '/');
	*pos = '\0';

	/* Use the official host name, not aliases */
	he = gethostbyname (hostname);
	assert (he != NULL);
	assert (he->h_name != NULL);
	strcpy (hostname, he->h_name);

	/* Add the port to hostname */
	pos = strchr (hostaddr, '/');
	strcat (hostname, pos);
	
	d->pid = pid;
	d->filename = strdup (filename);
	assert (d->filename !=NULL);
	d->hostaddr = strdup (hostaddr);
	assert (d->hostaddr !=NULL);
	n = chain_enqueue (fc_data.data, d);
	assert (n != NULL);
	return chain_length (fc_data.data);
}



/* -----------------------------------------------------------------
 * Function:    f1m_fc_remove_client_data ()
 *
 * Description: Remove data from the 'clients' database
 *
 * Returns:     Number of remaining clients
 */
PUBLIC int
f1m_fc_remove_client_data (int pid)
{
	chain_prune (fc_data.data,
				 f1m__fc_find_client_by_pid_and_delete, (void *)(&pid));
	return chain_length (fc_data.data);
}



/* -----------------------------------------------------------------
 * Function:    f1m_tp_add_servdata_tid ()
 *
 * Description: Add data to the 'tids' database
 *
 * Returns:     Pointer to the node with data or NULL on error
 */
PUBLIC serv_data_tid_t *
f1m_tp_add_servdata_tid (int gtid, int servid, int ltid, int rfd, int wfd)
{
	chain_node_t *n;
	serv_data_tid_t *d = malloc (sizeof (*d));
	assert (d != NULL);
	if (d != NULL) {
		d->gtid = gtid;
		d->servid = servid;
		d->ltid = ltid;
		d->fd[0] = rfd;
		d->fd[1] = wfd;
		n = chain_enqueue (tp_data.tids, d);
		assert (n != NULL);
	}
	return d;
}



/* -----------------------------------------------------------------
 * Function:    f1m_tp_sltid_2_tiddata ()
 *
 * Description: Get the server data ssociated with a pair (serverid, ltid)
 *
 * Returns:     Pointer to data
 */
PUBLIC serv_data_tid_t *
f1m_tp_sltid_2_tiddata (int servid, int ltid)
{
	int sltid[2];
	chain_node_t *n;

	sltid[0] = servid;
	sltid[1] = ltid;
	n = chain_find (tp_data.tids, &sltid, f1m__tp_find_by_sltid, NULL);
	if (n == NULL) return NULL;
	return (serv_data_tid_t *)chain_node_data (n);
}


/* -----------------------------------------------------------------
 * Function:    f1m_tp_add_servdata_machine ()
 *
 * Description: Add data to the 'machines' database
 *
 * Returns:     Pointer to the node with data or NULL on error
 */
PUBLIC serv_data_machine_t *
f1m_tp_add_servdata_machine (int servid, const char *machine, int rfd, int wfd)
{
	chain_node_t *n;
	serv_data_machine_t *d = malloc (sizeof (*d));
	assert (d != NULL);
	if (d != NULL) {
		d->servid = servid;
		d->machine = strdup (machine);
		d->fd[0] = rfd;
		d->fd[1] = wfd;
		/***/
		d->p_reps = chain_create();
		assert(d->p_reps);
		
		/* cria thread de leitura dos tkout's */
		assert(pthread_create(&(d->thid), NULL, f1m__tkout_reader, d)==0);
		pthread_detach(d->thid);
		assert(d->p_reps);
		/***/
		assert (d->machine != NULL);
		n = chain_enqueue (tp_data.machines, d);
		assert (n != NULL);
	}
	return d;
}


/* -----------------------------------------------------------------
 * Function:    f1m_tp_gtid_2_servdata ()
 *
 * Description: Get the server data ssociated with a global 'gtid' 
 *
 * Returns:     Pointer to data
 */
PUBLIC serv_data_tid_t *
f1m_tp_gtid_2_servdata (int gtid)
{
	chain_node_t *n = chain_find (tp_data.tids, &gtid,
								  f1m__tp_find_by_gtid, NULL);
	if (n == NULL) return NULL;
	return (serv_data_tid_t *)chain_node_data (n);
}


/* -----------------------------------------------------------------
 * Function:    f1m_tp_machine_2_servdata ()
 *
 * Description: Get the server data ssociated with a machine 
 *
 * Returns:     Pointer to data
 */
PUBLIC serv_data_machine_t *
f1m_tp_machine_2_servdata (const char *machine)
{
	chain_node_t *n = chain_find (tp_data.machines, machine,
								  f1m__tp_find_by_machine, NULL);
	if (n == NULL) return NULL;
	return (serv_data_machine_t *)chain_node_data (n);
}


/* -----------------------------------------------------------------
 * Function:    f1m_tp_servid_2_servdata ()
 *
 * Description: Get the server data ssociated with a server ID 
 *
 * Returns:     Pointer to data
 */

PUBLIC serv_data_machine_t *
f1m_tp_servid_2_servdata (int servid)
{
	chain_node_t *n;
	n = chain_find (tp_data.machines, &servid,
								  f1m__tp_find_by_servid, NULL);
	if (n == NULL) return NULL;
	return (serv_data_machine_t *)chain_node_data (n);
}


/* -----------------------------------------------------------------
 * Function:    f1m_tp_apply_to_all_machines ()
 *
 * Description: Request a service to all the machines
 *
 * Returns:     List of replies (tkout_t)
 */
PUBLIC chain_t *
f1m_tp_apply_to_all_machines (tkin_t *req)
{
	return chain_walk_collect (tp_data.machines, f1m__tp_apply_tkin, req);
}

/* onde colocar??? */

/* funcao de comparacao */
int f1m__tkout_id_cmp( void *data, void *args)
{
	if (((tkout_t *)data)->tid == (int)args)
		return 1;
	return 0;
		
}

int f1m__find_tid_notid( void *data, void *args)
{
	serv_data_tid_t *serv;
	int ltid = ((tkout_t *)data)->tid;
	int servid = (int)args;
	serv = f1m_tp_sltid_2_tiddata(servid, ltid);
	if (serv == NULL) 
		return 1;
	return 0;
}


tkout_t *f1m_read_tkout(int servid, int ltid)
{
	serv_data_machine_t *serv;
	tkout_t *resp;
	serv = f1m_tp_servid_2_servdata(servid);
	assert(serv);

	/* tratamento especial para "ltid=-1" - procura na fila por um i
	 * tid que nao exista na lista de tids deste servidor! 
	 */
	if (ltid == -1)
		resp = (tkout_t *)chain_dequeue_cond(serv->p_reps, f1m__find_tid_notid, (void *)servid);
	else {
		resp = (tkout_t *)chain_dequeue_cond(serv->p_reps, f1m__tkout_id_cmp, ltid);
		assert(ltid == resp->tid);
	}
	
	return resp;
}


void *f1m__tkout_reader(void *args)
{
	tkout_t *tk;
	chain_node_t *n;
	serv_data_machine_t *serv = (serv_data_machine_t *)args;
	assert(serv->p_reps);
	for (;;)
	{
		tk = f0m_read_tkout(serv->fd[0]);
		assert(tk!=NULL);
		n = chain_enqueue(serv->p_reps, tk);
		assert(n != NULL);
	}
	/* thread responsavel pela leitura dos tkout's */
	return NULL; /* blargh */
}

