/* *****************************************************************
 *
 *  $Id: f2m_lib.h,v 1.3 2002/08/06 22:03:34 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/06 15:27:33 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Prototypes of the services provided by FIDDLE-2
 *
 *		f2m_initialize ()		-- Initialize FIDDLE-2
 *		f2m_terminate ()		-- Terminate FIDDLE-2
 *		f2m_clients ()			-- List current clients data
 *		f2m_tids ()				-- List existing task IDs
 *		f2m_kill ()				-- Kill a target process
 *		f2m_attach ()			-- Attach to a running process
 *		f2m_detach ()			-- Detach from a running process
 *		f2m_symbol_file ()		-- Specify whre to find program symbols
 *		f2m_file ()				-- Load program and symbols into memory
 *		f2m_run ()				-- Run a program
 *		f2m_step ()				-- Executite until the next instruction
 *		f2m_next ()				-- Executite until the next instruction
 *		f2m_continue ()			-- Continue the execution
 *		f2m_finish ()			-- Execute until returning from
 *								   current stack frame
 *		f2m_signal ()			-- Send a signal to a process
 *		f2m_break ()			-- Set a breakpoint
 *		f2m_delete ()			-- Delete a breakpoint
 *		f2m_set_variable ()		-- Change the contents of a variable
 *		f2m_call ()				-- Call a function in the current
 *								   context of the target process
 *		f2m_evaluate ()			-- Evaluate an expression in the current
 *								   context of the target process
 *		f2m_display ()			-- Evaluate and display  an expression 
 *								   every time the process execution stops
 *		f2m_undisplay ()		-- Undisplay an expression
 *		f2m_list ()				-- List the process source code
 *		f1m_info_locals ()		-- Get name and value of local variables
 *		f1m_info_args ()		-- Get name and value of function argument
 *		f2m_info_line ()		-- Get info about current line
 *		f2m_info_program ()		-- Get info about current program
 *		f2m_info_stack ()		-- Get info about current stack frames
 *		f2m_up ()				-- Go up in the stack frame list
 *		f2m_down ()				-- Go down in the stack frame list
 *		f2m_frame ()			-- Select a specific stack frame
 *		f2m_info_break ()		-- Get info about current breakpoint
 *		f2m_info_display ()		-- Get info about current display expressions
 *		f2m_thread ()			-- Select a thread
 *		f2m_info_threads ()		-- Get info about the existing threads
 *		f2m_tty ()				-- Set a TTY for future IO
 *		f2m_sendto ()			-- Send a command directly to the debugger
 *		f2m_tkin ()				-- Send an "encoded" command
 *
 * -----------------------------------------------------------------
 */
#ifndef __F2_LIB_H
#define __F2_LIB_H


#include "glob.h"
#include "f0s_lib.h"
#include "f0m_tkin.h"


/* Flags to be used in f2m_initialize() */
#define START_SERVER	0x1
#define KEEP_SERVER		0x2


__BEGIN_DECLS
 
/* -----------------------------------------------------------------
 * Function:    int status = f2m_initialize (const char *host, 
 *											 const char *service,
 *											 int flags)
 *
 * Description: Initialize all the necessary data structures
 *				'host' is the name of the host where the main daemon
 *					   is running
 *				'service' is the service name (or decimal address)
 *				'flags' | START_SERVER if server should be started
 *						| KEEP_SERVER ifserver should be kept alive
 *
 * Returns:     The status
 */
int f2m_initialize (const char *host, const char *service, int flags);


/* -----------------------------------------------------------------
 * Function:    int status = f2m_terminate (int halt)
 *
 * Description: Terminate the debugging in this node.  Detach from all 
 *				processes and kil all CLDs.
 *
 * Returns:     The status
 */
int f2m_terminate (int halt);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_clients (const char *options)
 *
 * Description: Get a list of current Fiddle cleints.
 *				'options == "-m machine"' report only clients in <machine>
 *				'options == "-n nmae"' report only clients with <name>
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_clients (const char *options);


/* -----------------------------------------------------------------
 * Function:    f2m_add_debugger ()
 *
 * Description: Add a 'ready to be controlled' debugger to Fiddle
 *				'dbg_pid' should contain the PID of the debugger
 *				'tp_pid' should contain the PID of the target process
 *						 or -1 if process is not loaded or not running yet
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_add_debugger (const char *machine, int dbg_pid, int tp_pid,
						   const char *cmd_channel, const char *rep_channel);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_tids (const char *options)
 *
 * Description: Get a list of all existing process TIDs.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_tids (const char *options);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_kill (int tid)
 *
 * Description: Kill a process being debugger but let the CLD
 *				    available for a next usage
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_kill (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_attach (const char *machine, int tp_pid)
 *
 * Description: Find an unsused component-level debugger or start one
 *				if none was found.  Then attach the debugger to the
 *				process 'tp_pid'.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_attach (const char *machine, int tp_pid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_detach (int tid)
 *
 * Description: Detach a component-level debugger from the process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_detach (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_symbol_file (int tid, 
 *									        	const char *machine,
 *											    const char *file)
 *
 * Description: Load the symbols
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_symbol_file (int tid, const char *machine, const char *file);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_file (int tid,
 *										 const char *machine,
 *										 const char *file)
 *
 * Description: Load a process image into memory
 *				'tid == -1' load a new process image
 *				'tid != -1' reload a process image
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_file (int tid, const char *machine, const char *file);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_run (int tid, const char *args)
 *
 * Description: Start the execution of a process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_run (int tid, const char *args);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_step (int tid)
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps into the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_step (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_next (int tid)
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps over the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_next (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_continue (int tid)
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_continue (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_finish (int tid)
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_finish (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_signal (int tid, int signum)
 *
 * Description: Send a signal to a process.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_signal (int tid, int signum);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_break (int tid, const char *file,
 *									   int line, const char *function)
 *
 * Description: Set a breakpoint.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_break (int tid, const char *file, int line, const char *function);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_delete (int tid, int bpid)
 *
 * Description: Remove a breakpoint.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_delete (int tid, int bpid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_set_variable (int tid,
 *									   const char *variable,
 *									   const char *expression)
 *
 * Description: Set 'variable' to 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_set_variable (int tid, const char *variable,
						 const char *expression);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f2m_call (int tid,
 *									   const char *expression)
 *
 * Description: Call a function and get it's return value
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_call (int tid, const char *expression);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_evaluate (int tid,
 *									   const char *expression)
 *
 * Description: Evaluate an 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_evaluate (int tid, const char *expression);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_display (int tid,
 *									     const char *expression)
 *
 * Description: Display 'expression' every time the program stops
 *				If 'expression == NULL' list list all expressions/values
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_display (int tid, const char *expression);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_undisplay (int tid, int dispid)
 *
 * Description: Remove a display expression
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_undisplay (int tid, int dispid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_list (int tid, const char *file,
 *									   int start_line,
 *									   const char *function,
 *									   int n_lines)
 *
 * Description: List a source file.  
 *              The arguments 'start_line' and 'function' are exclusive.
 *              The arguments 'n_lines' is optional.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_list (int tid, const char *file, int start_line,
				 const char *function, int n_lines);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_info_locals (int tid)
 *
 * Description: Get name and value of local variables
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_locals (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_info_args (int tid)
 *
 * Description: Get name and value of current function arguments
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_args (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_info_line (int tid, const char *file,
 *									        int line,
 *									        const char *function)
 *
 * Description: Get information about a source line.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_line (int tid, const char *file, int line,
					  const char *function);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f2m_info_program (int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_program (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_info_stack (int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_stack (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_up (int tid)
 *
 * Description: Go "up" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_up (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_down (int tid)
 *
 * Description: Go "down" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_down (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_frame (int tid, int frame)
 *
 * Description: Select a specific stack frame
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_frame (int tid, int frame);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_info_break (int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_break (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_info_display (int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_display (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_thread (int tid, int threadID)
 *
 * Description: Switch the current thread.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_thread (int tid, int threadID);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_info_threads (int tid)
 *
 * Description: Get information about the existing threads.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_info_threads (int tid);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f2m_tty (int tid,
 *									  const char *file)
 *
 * Description: Set the I/O tty for the process being debugged
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_tty (int tid, const char *file);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_sendto (int tid, const char *comand)
 *
 * Description: Send a command string directly to the debugger.
 *				'tkout' contains the output (NOT PARSED).
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_sendto (int tid, const char *command);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f2m_tkin3 (int tid, tkin_t *tk, int delete)
 *
 * Description: Set a command already packed into a 'tkin_t' strucure
 *				to process 'tid'.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f2m_tkin3 (int tid, tkin_t *tk, int delete);


__END_DECLS

#endif /* __F2_LIB_H */
