/* *****************************************************************
 *
 *  $Id: f2m_queues.c,v 1.3 2002/09/27 12:43:03 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/26 13:27:55 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implementation for the enqueue requests and dequeue
 *				 replies services
 *
 *		f0m_queues_init ()		-- Initialization
 *		f0m_queues_terminate ()	-- Termination
 *		f0m_enqueue_request ()	-- Enqueue a request
 *		f0m_dequeue_request ()	-- Dequeue a request
 *		f0m_enqueue_reply ()		-- Enqueue a reply
 *		f0m_dequeue_reply ()		-- Dequeue a reply
 *
 * -----------------------------------------------------------------
 */

#include "config.h"

#include <stdlib.h>
#include <assert.h>
#include <pthread.h>

#include "chain.h"

#include "f0s_tkout.h"

#include "f2m_queues.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

PRIVATE chain_t *req_q = NULL;
PRIVATE chain_t *rep_q = NULL;


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    reqid_t reqid = generate_new_request_id (void)
 *
 * Description: Generate a new request ID
 *
 * Returns:     The new request ID
 */
static reqid_t generate_new_request_id (void)
{
	static pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
	static reqid_t the_id = 0;
	reqid_t reqid ;
	
	pthread_mutex_lock (&mut);
	reqid = ++the_id;
	pthread_mutex_unlock (&mut);
	return reqid;
}


/* -----------------------------------------------------------------
 * Function:    int cmp_id_f (void *data, void *args)
 *
 * Description: Test 'if (((reply_t *)data)->id == (int) args)'
 *
 * Returns:     1 if condition evaluates to TRUE
 *				0 otherwise
 */
static int cmp_id_f (void *data, void *args)
{
	assert (data != NULL);
	if (((reply_t *)data)->id == *((int *)args))
		return 1;				/* TRUE */
	return 0;					/* FALSE */
}


/* -----------------------------------------------------------------
 * Function:    int cmp_one_f (void *data, void *args)
 *
 * Description: Test 'if (((reply_t *)data)->id == (int) args)'
 *
 * Returns:     1 if condition evaluates to TRUE
 *				0 otherwise
 */
static int cmp_one_f (void *data, void *args)
{
	assert (data != NULL);
	return 1;				/* TRUE */
}


/* -----------------------------------------------------------------
 * Function:    request_t *req = tkin_2_request (const tkin_t *tk)
 *
 * Description: Generate a request ID and pack it with the 'tkin_t'
 *				token into a 'request_t' structure.
 *
 * Returns:     A pointer to a newly allocated 'request_t' structure
 */
static request_t *tkin_2_request (const tkin_t *tk)
{
	request_t *req = malloc (sizeof (*req));
	assert (req != NULL);
	req->id = generate_new_request_id ();
	req->tk = tk;
	return req;
}


/* -----------------------------------------------------------------
 * Function:    reply_t *rep = tkout_2_reply (int servid, int id, 
 *											  const tkout_t *tk)
 *
 * Description: Pack the request ID, task ID and the reply token into
 *				a 'reply_t' structure.
 *
 * Returns:     A pointer to a newly allocated 'reply_t' structure
 */
static reply_t *tkout_2_reply (int id, const tkout_t *tk)
{
	reply_t *rep = malloc (sizeof (*rep));
	assert (rep != NULL);
	rep->id = id;
	rep->tk = tk;
	return rep;
}




/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    int f2m_queues_init (void)
 *
 * Description: Initialize queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
int f2m_queues_init (void)
{
	assert (req_q == NULL);
	assert (rep_q == NULL);
	req_q = chain_create ();
	rep_q = chain_create ();
	assert (req_q != NULL);
	assert (rep_q != NULL);
	return 0;
}

/* -----------------------------------------------------------------
 * Function:    int f2m_queues_terminate (void)
 *
 * Description: Terminate queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
int f2m_queues_terminate (int halt)
{
	if (halt) {
		assert (req_q != NULL);
		assert (rep_q != NULL);
		chain_destroy (req_q);
		chain_destroy (rep_q);
	}
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    reqid_t f2m_enqueue_request (const tkin_t *req)
 *
 * Description: Enqueue a request of a service
 *				'req'	 Token with the request code and arguments
 *
 * Returns:     The request ID
 */
reqid_t f2m_enqueue_request (const tkin_t *tk)
{
	chain_node_t *status;
	/* Enqueue request */
	request_t *req = tkin_2_request (tk);
	assert (req_q != NULL);
	status = chain_enqueue (req_q, req);
	assert (status != NULL);
	return req->id;
}


/* -----------------------------------------------------------------
 * Function:    request_t *req f2m_dequeue_request (void)
 *
 * Description: Get next entry from request queue
 *
 * Returns:     A structure 'request_t'
 */
request_t *f2m_dequeue_request (void)
{
	request_t *req;

	assert (req_q != NULL);
	req = (request_t *) chain_dequeue_cond (req_q, cmp_one_f, NULL);
	return req;
}


/* -----------------------------------------------------------------
 * Function:    int repid = f2m_enqueue_reply (reqid_t id,
 *											   const tkout_t *rep)
 *
 * Description: Enqueue a reply to a service
 *				'id'	The request ID
 *				'rep'	Token with the reply code and arguments
 *
 * Returns:     The status
 */
int f2m_enqueue_reply (reqid_t id, const tkout_t *tk)
{
	reply_t *rep;
	void *err;

	rep = tkout_2_reply (id, tk);
	assert (rep_q != NULL);
	err = chain_enqueue (rep_q, rep);
	assert (err != NULL);
	return 0;
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *tk = f2m_dequeue_reply (reqid_t reqid)
 *
 * Description: Get the reply to the request with ID 'reqid'
 *
 * Returns:     Pointer to the reply ('tkout_t') structure
 */
tkout_t *f2m_dequeue_reply (reqid_t reqid)
{
	reply_t *rep;				/* The reply structure */

	assert (rep_q != NULL);
	rep = (reply_t *) chain_dequeue_cond (rep_q, cmp_id_f, &reqid);
	return (tkout_t *)rep->tk;
}

