/* *****************************************************************
 *
 *  $Id: Fiddle_j.c,v 1.1 2002/08/31 15:23:00 rja Exp $
 *
 *  Copyright (c) 2000 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/31 13:43:19 rja>
 *
 *  Ricardo Anastacio [jml] <rja@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *    C implementation of the JAVA methods that call
 *  the fiddle engine.
 * -----------------------------------------------------------------
 */


/* #################################################################
 *      Defines, Types and Global Variables
 * ################################################################# */

#include "Fiddle_j.h"

#include <string.h>
#include <fiddle/fx_lib.h>
#include "common.h"
#include "do_err.h"
#include "do_pos.h"
#include "do_OoB.h"
#include "do_rep.h"


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

#ifdef EBUG
PRIVATE void passo(float f) {
  fprintf(stderr, "Passo %.2f.\n", f);
  fflush(stderr);
}
#endif


/*
 * Checks the tkout_t structure for common errors throwing an
 * exception if an error is found. A negative value is returned if
 * normal checking failed. Otherwise a non-negative value is returned.
 * (+) env, the JAVA enviroment variable;
 * (+) tk, the tkout_t structure to validate;
 * (-)  1, An exception has been thrown. Exiting;
 *      0,  No errors found;
 *     -1, Unable to find an Exception in a memory error;
 */
PRIVATE int check_for_errors(JNIEnv *env, tkout_t *tk) {
  jclass     ExcCls;
  jthrowable ExcObj;
  char       errMsg[MAX_STR];
  
#ifdef EBUG
	passo(3.1);
#endif
  
  if(tk == NULL) {
	  strcpy(errMsg, "Service returned an empty tkout_t structure. Memory Error!");
	  ExcCls = (*env)->FindClass(env, "java/lang/NullPointerException");
	  if(ExcCls == 0)	return -1;  /* Unable to find an Exception to throw. Giving up! */
	  (*env)->ThrowNew(env, ExcCls, errMsg);
	  return 1;
  }
  
#ifdef EBUG
    passo(3.2);
#endif
  
	if(tk->err != NULL) {
#ifdef f2m
	  fprintf(stderr, "fiddle: f2mj: error code %d - %s\n", tk->err->code, tk->err->str);
#else	  
	  fprintf(stderr, "fiddle: f1mj: error code %d - %s\n", tk->err->code, tk->err->str);
#endif

// Fill the exception to be thrown and throw it
	  ExcObj = do_err(env, tk);
	  (*env)->Throw(env, ExcObj);
	  return 1;
  }

#ifdef EBUG
	passo(3.3);
#endif
  
  if(tk->status > FIDDLE_OK) {
	  strcpy(errMsg, "Status different of FIDDLE_OK.");

#ifdef EBUG
	  passo(3.4);
#endif

// Fill the exception to be thrown and throw it
	  ExcObj = do_status(env, tk);
	  (*env)->Throw(env, ExcObj);
	  return 1;
  }
  
#ifdef EBUG
	passo(3.5);
#endif
  
  return 0;
}



/*
 * Processes independent structures of the tkout_t object. That is, it
 * processes the structures that always contains the same
 * independently of which function is called.
 * (+) env, the JAVA enviroment variable;
 * (+/-) iObj, a structure holding the independent objects;
 * (+) tk, the tkout_t structure to process;
 * (-) returns a negative value if by any reason the processing could
 *     not ocurr; 0, otherwise;
 */
PRIVATE int process_independent_structures(JNIEnv *env, indOb_t iObj,\
										   tkout_t *tk) {

  jclass    cls;
  jmethodID mid;
  
#ifdef EBUG
  passo(6.1);
#endif
  
// Simple, direct members
  if(tk->str==NULL) tk->str = strdup("");
  iObj->tid = tk->tid;  // Symbolic Task IDentifier;
  iObj->str = (*env)->NewStringUTF(env, tk->str); // Back-end debugger's output;

#ifdef EBUG
  passo(6.2);
#endif
  
// Not so simple or direct filling members

  // status object from Tkout_j
  cls = (*env)->FindClass(env, CODE_J_SIG);

#ifdef EBUG
  passo(6.3);
#endif

  if(cls == 0) return -1;
#ifdef EBUG
  passo(6.4);
#endif
  
  mid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");
#ifdef EBUG
  passo(6.5);
#endif
  
  if(mid == 0) return -1;
#ifdef EBUG
  passo(6.6);
#endif
  
  iObj->status = (*env)->NewObject(env, cls, mid, tk->status);

#ifdef EBUG
  passo(6.7);
#endif

  // OoB object from Tkout_j
  process_OoB(env, iObj, tk);
  
#ifdef EBUG
  passo(6.8);
#endif

// Pos object from Tkout_j
  process_pos(env, iObj, tk);
	  
#ifdef EBUG
  passo(6.9);
#endif
	
  return 0;
}






/*
 * Code common to all te following public functions.
 * The only mutable things are the names of functions
 * passed as arguments to this one.
 * (+) env, the JAVA enviroment;
 * (+) tk, the fiddle function reply;
 * (+) process_rep_function, the function to process the
 *     fiddle function reply;
 * (-) o objecto Tkout_j j construdo e inicializado ou
 *     NULL caso uma excepo tenha sido lanada.
 */
PRIVATE jobject createTkout_j(JNIEnv *env, tkout_t *tk, int (*process_rep_function)(void *, void *)) {
  jclass    ExcCls;
  jobject   jtk;
  indOb_t   iObj;
  repOb_t   rObj;
  char      errMsg[MAX_STR];
  int       res;
  
  res = check_for_errors(env, tk);
  if(res == 1) return NULL; // An exception has been thrown!
  if(res != 0) {
// Checking for errors returned error without throwing an exception.	
// Bad... Really bad. Trying to throw a generic exception.
	strcpy(errMsg, "Checking for errors ended abruptly. Unknown error!");
	ExcCls = (*env)->FindClass(env, "Ljava/lang/Exception;");
	(*env)->ThrowNew(env, ExcCls, errMsg);
	return NULL;
  }

#ifdef EBUG
  passo(4);
#endif
  
// Alocate memory space for the independent object
  iObj = (indOb_t) malloc(sizeof(_indOb_t));

#ifdef EBUG
  passo(5);
#endif
  
// Alocate memory space for the reply object
  rObj = (repOb_t) malloc(sizeof(_repOb_t));
  

#ifdef EBUG
  passo(6);
#endif
  
// Processing status, tid, pos, oob & str
  if(process_independent_structures(env, iObj, tk) != 0) {
	strcpy(errMsg, "Error trying to process independent structures.");
	ExcCls = (*env)->FindClass(env, "Ljava/lang/Exception;");
	(*env)->ThrowNew(env, ExcCls, errMsg);
	return NULL;
  }

#ifdef EBUG
  passo(7);
#endif
  
// Processing reply  
  process_reply(env, rObj, tk, process_rep_function);

#ifdef EBUG
  passo(8);
#endif
  
// and finally initializing the JAVA Tkout_j object with its arguments
  jtk = initTkout_j(env, iObj, rObj);
  if((*env)->ExceptionOccurred(env)!=NULL) {
	(*env)->ExceptionDescribe(env);
	return NULL;
	// (*env)->ExceptionClear(env);
  }
  
#ifdef EBUG
  fprintf(stderr, "jtk=0x%x", (unsigned int) jtk);
  passo(9);
#endif
  
// Free the reply object alocated memory space
  free(rObj);
	
#ifdef EBUG
  passo(10);
#endif
  
// Free the independent object alocated memory space	
  free(iObj);
	
#ifdef EBUG
  passo(11);
#endif

  f0s_tkout_delete(tk);

#ifdef EBUG
  passo(12);
#endif
  
  return jtk;
}
 

/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


/*
 * Class:     Tkout_0005fj
 * Method:    initialize
 * Signature: (Ljava/lang/String;Ljava/lang/String;I)I
 */
JNIEXPORT jint JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_initialize(JNIEnv *env, jobject this, jstring jhost, jstring jservice, jint flags) {

  int retVal;
#ifdef f2m
// Converting the String to a char *
  const char *host, *service;
  
// Converting the Strings to char *
  if(jhost != NULL) {
	host = (*env)->GetStringUTFChars(env, jhost, 0);
	if(*host=='\0') {(*env)->ReleaseStringUTFChars(env, jhost, host); host=NULL;}
  } else host = NULL;
  if(jservice != NULL) {
	service = (*env)->GetStringUTFChars(env, jservice, 0);
	if(*service=='\0') {(*env)->ReleaseStringUTFChars(env, jservice, service); service=NULL;}
  } else service = NULL;

#ifdef EBUG
	printf("Vou fazer fx_initialize com os argumentos: host=|%s|, service=|%s|, flags=|%d|\n", host, service, (int) flags);
#endif

	retVal = fx_initialize("localhost", "12345", (int) flags);
	
// We don't need the char * anymore, let's release it.
	if(host!=NULL)
	  (*env)->ReleaseStringUTFChars(env, jhost, host);
	if(service!=NULL)
	  (*env)->ReleaseStringUTFChars(env, jservice, service);
#else

#ifdef EBUG
	printf("Vou fazer fx_initialize\n");
#endif

    retVal = fx_initialize();
#endif
	
#ifdef EBUG
	printf("Fez o fx_initialize retornando %d\n", retVal);
#endif

	return retVal;
}





/*
 * Class:     Fiddle_0005fj
 * Method:    die
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_die(JNIEnv *env, jobject this, jint flag_halt) {
  int res;

#ifdef f2m
  res = fx_terminate((int) flag_halt);
#else
  res = fx_terminate();
#endif
  
#ifdef EBUG
  if(res != 0) {
	fprintf(stderr, "Fiddle not terminated. fx_terminate returned %d.\nCleaning the JAVA object anyway.\n", res);
	fflush(stderr);
  }
#endif

}






/*
 * Class:     Fiddle_0005fj
 * Method:    clients
 * Signature: (Ljava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_clients(JNIEnv *env, jobject this, jstring joptions) {
  jclass     ExcCls;
  char       errMsg[MAX_STR];
  
#ifdef f2m

  tkout_t    *tk;
	
// Converting the String to a char *
  const char *options = (*env)->GetStringUTFChars(env, joptions, 0);
	
#ifdef EBUG
  fprintf(stderr, "Doing clients with args: options=|%s|\n", options);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_clients(options);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, joptions, options);
  
#ifdef EBUG
  passo(3);
#endif

  return createTkout_j(env, tk, process_rep_clients);

#else
  	strcpy(errMsg, "Fiddle_j: native lib reported service unavailable");
	ExcCls = (*env)->FindClass(env, SERVICENOTAVAILABLEEXCEPTION_SIG);
	(*env)->ThrowNew(env, ExcCls, errMsg);
	return NULL;
#endif

}






/*
 * Class:     Tkout_0005fj
 * Method:    file
 * Signature: (Ljava/lang/String;Ljava/lang/String;)V
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_file(JNIEnv *env, jobject this, jstring jmachine, jstring jfile) {

  tkout_t   *tk;
  
// Converting the String to a char *
  const char *machine = (*env)->GetStringUTFChars(env, jmachine, 0);
  const char *file = (*env)->GetStringUTFChars(env, jfile, 0);

#ifdef EBUG
  fprintf(stderr, "Doing file with args: machine=|%s|, file=|%s|\n", machine, file);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_file(-1, machine, file);
  
#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jmachine, machine);
  (*env)->ReleaseStringUTFChars(env, jfile, file);
  
#ifdef EBUG
  passo(3);
#endif

  return createTkout_j(env, tk, process_rep_file);
}



/*
 * Class:     Fiddle_0005fj
 * Method:    setBreakpoint
 * Signature: (ILjava/lang/String;ILjava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_setBreakpoint(JNIEnv *env, jobject this,
													   jint tid, jstring jfile,
													   jint line, jstring jfunction) {
  tkout_t   *tk;
  const char *function, *file;

// Converting the String to a char *
  if(jfunction != NULL) {
	function = (*env)->GetStringUTFChars(env, jfunction, 0);
	if(*function=='\0') {(*env)->ReleaseStringUTFChars(env, jfunction, function); function=NULL;}
  } else function = NULL;
  if(jfile != NULL) {
	file = (*env)->GetStringUTFChars(env, jfile, 0);
	if(*file=='\0') {(*env)->ReleaseStringUTFChars(env, jfile, file); file=NULL;}
  } else file = NULL;

#ifdef EBUG
  fprintf(stderr, "Doing break with args: tid=%d, file=%s, line=%d, function=", (int) tid, file, (int) line);
  if(function==NULL) fprintf(stderr, "(null)\n");
  else fprintf(stderr, "%s\n", function);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_break(tid, file, line, function);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  if(function!=NULL) (*env)->ReleaseStringUTFChars(env, jfunction, function);
  if(file!=NULL) (*env)->ReleaseStringUTFChars(env, jfile, file);
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_break);
}




/*
 * Class:     Fiddle_0005fj
 * Method:    run
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_run(JNIEnv *env, jobject this,
												jint tid, jstring jargs) {
  tkout_t   *tk;
  const char *args;

// Converting the String to a char *
  if(jargs != NULL) {
	args = (*env)->GetStringUTFChars(env, jargs, 0);
	if(*args=='\0') {(*env)->ReleaseStringUTFChars(env, jargs, args); args=NULL;}
  } else args = NULL;

#ifdef EBUG
  fprintf(stderr, "Doing run with args: tid=%d\n", (int) tid);
  passo(1);
#endif
  
// the C fiddle function!!!

  tk = fx_run(tid, args);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  if(args!=NULL) (*env)->ReleaseStringUTFChars(env, jargs, args);
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_run);
}






/*
 * Class:     Fiddle_0005fj
 * Method:    continueExecution
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_continueExecution(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
 
#ifdef EBUG
  fprintf(stderr, "Doing continue with args: tid=%d\n", (int) tid);
  passo(1);
#endif
  
// the C fiddle function!!!

  tk = fx_continue(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.

#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_continue);
}










/*
 * Class:     Fiddle_0005fj
 * Method:    delete
 * Signature: (II)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_delete(JNIEnv *env, jobject this, jint tid, jint bpid) {
  tkout_t   *tk;

#ifdef EBUG
  fprintf(stderr, "Doing delete with args: tid=%d, bpid=%d\n", (int) tid, (int) bpid);
  passo(1);
#endif
  
// the C fiddle function!!!

  tk = fx_delete(tid, bpid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_delete);
}





/*
 * Class:     Fiddle_0005fj
 * Method:    list
 * Signature: (ILjava/lang/String;ILjava/lang/String;I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_list(JNIEnv *env, jobject this,
											  jint tid, jstring jfile,
											  jint line, jstring jfunction,
											  jint n) {
  tkout_t   *tk;
  const char *file, *function;
  
// Converting the Strings to char *
  if(jfile != NULL) {
	file = (*env)->GetStringUTFChars(env, jfile, 0);
	if(*file=='\0') {(*env)->ReleaseStringUTFChars(env, jfile, file); file=NULL;}
  } else file = NULL;

  if(jfunction != NULL) {
	function = (*env)->GetStringUTFChars(env, jfunction, 0);
	if(*function=='\0') {(*env)->ReleaseStringUTFChars(env, jfunction, function); function=NULL;}
  } else function = NULL;
  
#ifdef EBUG
  fprintf(stderr, "Doing list with args: tid=%d, file=%s, line=%d, function=%s, n=%d\n", (int) tid, file, (int) line, function, (int) n);
  passo(1);
#endif


  
// the C fiddle function!!!

  tk = fx_list(tid, file, line, function, n);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  if(file!=NULL) (*env)->ReleaseStringUTFChars(env, jfile, file);
  if(function!=NULL) (*env)->ReleaseStringUTFChars(env, jfunction, function);
  
#ifdef EBUG
  passo(3);
#endif

  return createTkout_j(env, tk, process_rep_list);
}








/*
 * Class:     Fiddle_0005fj
 * Method:    tids
 * Signature: ()LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_tids(JNIEnv *env, jobject this) {
  tkout_t   *tk;

#ifdef EBUG
  fprintf(stderr, "Doing tids\n");
  passo(1);
#endif
  
// the C fiddle function!!!

  tk = fx_tids(NULL);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  // ** No chars to release ** 

  
#ifdef EBUG
  passo(3);
#endif

  return createTkout_j(env, tk, process_rep_tids);
}








/*
 * Class:     Fiddle_0005fj
 * Method:    attach
 * Signature: (Ljava/lang/String;I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_attach(JNIEnv *env, jobject this,
												jstring jmachine, jint tp_pid) {

  tkout_t   *tk;
  
// Converting the String to a char *
  const char *machine = (*env)->GetStringUTFChars(env, jmachine, 0);

#ifdef EBUG
  fprintf(stderr, "Doing attach with args: machine=|%s|, tp_pid=|%d|\n", machine, (int) tp_pid);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_attach(machine, tp_pid);
  
#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jmachine, machine);
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_attach);
}




/*
 * Class:     Fiddle_0005fj
 * Method:    step
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_step(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;

// Converting the String to a char *

#ifdef EBUG
  fprintf(stderr, "Doing step with args: tid=%d\n", (int) tid);
  passo(1);
#endif
  
// the C fiddle function!!!

  tk = fx_step(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_step);
}







/*
 * Class:     Fiddle_0005fj
 * Method:    next
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_next(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;

// Converting the String to a char *

#ifdef EBUG
  fprintf(stderr, "Doing next with args: tid=%d\n", (int) tid);
  passo(1);
#endif
  
// the C fiddle function!!!

  tk = fx_next(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_next);
}






/*
 * Class:     Fiddle_0005fj
 * Method:    finish
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_finish(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;

// Converting the String to a char *

#ifdef EBUG
  fprintf(stderr, "Doing finish with args: tid=%d\n", (int) tid);
  passo(1);
#endif
  
// the C fiddle function!!!

  tk = fx_finish(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release them.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_finish);
}











/*
 * Class:     Fiddle_0005fj
 * Method:    sendTo
 * Signature: ()LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_sendTo(JNIEnv *env, jobject this, jint tid, jstring jcmd) {
  tkout_t   *tk;
  
// Converting the String to a char *
  const char *cmd = (*env)->GetStringUTFChars(env, jcmd, 0);

#ifdef EBUG
  fprintf(stderr, "Doing sendto with args: tid=|%d|, cmd=|%s|\n", (int) tid, cmd);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_sendto(tid, cmd);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jcmd, cmd);
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_sendto);
}




/*
 * Class:     Fiddle_0005fj
 * Method:    display
 * Signature: (ILjava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_display(JNIEnv *env, jobject this, jint tid, jstring jexp) {
  tkout_t   *tk;
  
// Converting the String to a char *
  const char *exp = (*env)->GetStringUTFChars(env, jexp, 0);

#ifdef EBUG
  fprintf(stderr, "Doing display with args: tid=|%d|, expression=|%s|\n", (int) tid, exp);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_display(tid, exp);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jexp, exp);
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_display);
}




/*
 * Class:     Fiddle_0005fj
 * Method:    undisplay
 * Signature: (II)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_undisplay(JNIEnv *env, jobject this, jint tid, jint dispid) {
  tkout_t   *tk;
  
#ifdef EBUG
  fprintf(stderr, "Doing display with args: tid=|%d|, dispid=|%d|\n", (int) tid, (int) dispid);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_undisplay(tid, dispid);

#ifdef EBUG
  passo(2);
#endif
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_undisplay);
}




/*
 * Class:     Fiddle_0005fj
 * Method:    set_variable
 * Signature: (ILjava/lang/String;Ljava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_set_1variable(JNIEnv *env, jobject this, jint tid, jstring jvar, jstring jexp) {
  tkout_t   *tk;
  
// Converting the Strings to char *
  const char *var = (*env)->GetStringUTFChars(env, jvar, 0);
  const char *exp = (*env)->GetStringUTFChars(env, jexp, 0);

#ifdef EBUG
  fprintf(stderr, "Doing set variable with args: tid=|%d|, variable=|%s|, expression=|%s|\n", (int) tid, var, exp);
  passo(1);
#endif
  
// the C fiddle function!!!
  tk = fx_set_variable(tid, var, exp);

#ifdef EBUG
  passo(2);
#endif
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_set_variable);
}





/*
 * Class:     Fiddle_0005fj
 * Method:    evaluate
 * Signature: (ILjava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_evaluate(JNIEnv *env, jobject this, jint tid, jstring jexp) {
  tkout_t   *tk;
  
// Converting the Strings to char *
  const char *exp = (*env)->GetStringUTFChars(env, jexp, 0);

#ifdef EBUG
  fprintf(stderr, "Doing evaluate expression with args: tid=|%d|, expression=|%s|\n", (int) tid, exp);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_evaluate(tid, exp);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jexp, exp);
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_evaluate);
}




/*
 * Class:     Fiddle_0005fj
 * Method:    thread
 * Signature: (II)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_thread(JNIEnv *env, jobject this, jint tid, jint threadID) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing thread with args: tid=|%d|, threadID=|%d|\n", (int) tid, threadID);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_thread(tid, threadID);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_thread);
}






/*
 * Class:     Fiddle_0005fj
 * Method:    detach
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_detach(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing detach with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_detach(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_detach);
}


/*
 * Class:     Fiddle_0005fj
 * Method:    kill
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_kill(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing kill with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_kill(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_kill);
}



/*
 * Class:     Fiddle_0005fj
 * Method:    symbol_file
 * Signature: (ILjava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_symbol_1file(JNIEnv *env, jobject this, jint tid, jstring jmachine, jstring jfilename) {
  tkout_t   *tk;
  
// Converting the Strings to char *
  const char *machine = (*env)->GetStringUTFChars(env, jmachine, 0);
  const char *filename = (*env)->GetStringUTFChars(env, jfilename, 0);

#ifdef EBUG
  fprintf(stderr, "Doing symbol_file with args: tid=|%d|, machine=|%s|, filename=|%s|\n", (int) tid, machine, filename);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_symbol_file(tid, machine, filename);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jfilename, filename);
  (*env)->ReleaseStringUTFChars(env, jmachine, machine);
  
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_symbol_file);
}





/*
 * Class:     Fiddle_0005fj
 * Method:    signal
 * Signature: (II)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_signal(JNIEnv *env, jobject this, jint tid, jint signum) {
  tkout_t   *tk;
  
// Converting the Strings to char *

  
#ifdef EBUG
  fprintf(stderr, "Doing signal with args: tid=|%d|, signum=|%d|\n", (int) tid, (int) signum);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_signal(tid, signum);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.

  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_signal);
}





/*
 * Class:     Fiddle_0005fj
 * Method:    call
 * Signature: (ILjava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_call(JNIEnv *env, jobject this, jint tid, jstring jexpression) {
  tkout_t   *tk;
  
// Converting the Strings to char *
  const char *expression = (*env)->GetStringUTFChars(env, jexpression, 0);

#ifdef EBUG
  fprintf(stderr, "Doing call with args: tid=|%d|, expression=|%s|\n", (int) tid, expression);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_call(tid, expression);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jexpression, expression);
  
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_call);
}

/*
 * Class:     Fiddle_0005fj
 * Method:    info_line
 * Signature: (ILjava/lang/String;ILjava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_info_1line(JNIEnv *env, jobject this, jint tid, jstring jfile, jint line, jstring jfunction) {
  tkout_t   *tk;
  const char *file, *function;
  
// Converting the Strings to char *
  if(jfunction != NULL) {
	function = (*env)->GetStringUTFChars(env, jfunction, 0);
	if(*function=='\0') {(*env)->ReleaseStringUTFChars(env, jfunction, function); function=NULL;}
  } else function = NULL;
  if(jfile != NULL) {
	file = (*env)->GetStringUTFChars(env, jfile, 0);
	if(*file=='\0') {(*env)->ReleaseStringUTFChars(env, jfile, file); file=NULL;}
  } else file = NULL;

#ifdef EBUG
  fprintf(stderr, "Doing info_line with args: tid=|%d|, file=|%s|, line=|%d|, function=|%s|\n", (int) tid, file, (int) line, function);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_info_line(tid, file, line, function);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jfile, file);
  (*env)->ReleaseStringUTFChars(env, jfunction, function);
  
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_info_line);
}

/*
 * Class:     Fiddle_0005fj
 * Method:    info_program
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_info_1program(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing info_program with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_info_program(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_info_program);
}

/*
 * Class:     Fiddle_0005fj
 * Method:    info_stack
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_info_1stack(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing info_stack with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_info_stack(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_info_stack);
}

/*
 * Class:     Fiddle_0005fj
 * Method:    up
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_up(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing up with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_up(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_up);
}

/*
 * Class:     Fiddle_0005fj
 * Method:    down
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_down(JNIEnv *env, jobject this, jint tid) {
    tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing down with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_down(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_down);
}


/*
 * Class:     Fiddle_0005fj
 * Method:    frame
 * Signature: (II)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_frame(JNIEnv *env, jobject this, jint tid, jint frame) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing frame with args: tid=|%d|, frame=|%d|\n", (int) tid, (int) frame);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_frame(tid, frame);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_frame);
}


/*
 * Class:     Fiddle_0005fj
 * Method:    info_break
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_info_1break(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing info_break with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_info_break(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_info_break);
}

/*
 * Class:     Fiddle_0005fj
 * Method:    info_display
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_info_1display(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing info_display with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_info_display(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_info_display);
}


/*
 * Class:     Fiddle_0005fj
 * Method:    info_threads
 * Signature: (I)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_info_1threads(JNIEnv *env, jobject this, jint tid) {
  tkout_t   *tk;
  
// Converting the Strings to char *

#ifdef EBUG
  fprintf(stderr, "Doing info_threads with args: tid=|%d|\n", (int) tid);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_info_threads(tid);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_info_threads);

}


/*
 * Class:     Fiddle_0005fj
 * Method:    tty
 * Signature: (ILjava/lang/String;)LTkout_j;
 */
JNIEXPORT jobject JNICALL Java_unl_fct_di_fiddle_1j_Fiddle_1j_tty(JNIEnv *env, jobject this, jint tid, jstring jtty_dev) {
  tkout_t   *tk;
  const char *tty_dev;
  
// Converting the Strings to char *
  tty_dev = (*env)->GetStringUTFChars(env, jtty_dev, 0);

#ifdef EBUG
  fprintf(stderr, "Doing tty with args: tid=|%d|, tty_dev=|%s|\n", (int) tid, tty_dev);
  passo(1);
#endif

// the C fiddle function!!!
  tk = fx_tty(tid, tty_dev);

#ifdef EBUG
  passo(2);
#endif
  
// We don't need the char * anymore, let's release it.
  (*env)->ReleaseStringUTFChars(env, jtty_dev, tty_dev);
  
  
#ifdef EBUG
  passo(3);
#endif
  
  return createTkout_j(env, tk, process_rep_tty);
}
