/* *****************************************************************
 *
 *  $Id: chain.h,v 1.2 2002/08/14 01:30:25 jml Exp $
 *
 *  Copywright (c) 1999 by Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *  Time-stamp: <02/08/12 17:15:36 jml>
 *
 *  Universidade Nova de Lisboa
 *  Faculdade de Ciencias e Tecnologia
 *  Departamento de Informatica
 *  Parallel and Distributed Processing Group
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */
#ifndef __CHAIN_H
#define __CHAIN_H

#include "glob.h"



typedef int (*chain_walker_f) (void *data, void *args);
typedef void * (*chain_walker_collect_f) (void *data, void *args);
typedef int (*chain_compare_f) (const void *data, const void *what, void *args);


typedef struct chain_collect_s chain_collect_t;
typedef struct chain_node_s chain_node_t;
typedef struct chain_s chain_t;



/* -----------------------------------------------------------------
 * Function:    void *chain_data (const chain_node_t *node)
 *
 * Description: Get the data kept in the chain node
 *
 * Returns:     Pointer to the data in the node
 */
void *chain_node_data (const chain_node_t *node);


/* -----------------------------------------------------------------
 * Function:    chain_t *chain_create (void)
 *
 * Description: Create a new empty chain
 *
 * Returns:     Pointer to new chain
 */
chain_t *chain_create (void);


/* -----------------------------------------------------------------
 * Function:    void chain_destroy (chain_t *chain)
 *
 * Description: Delete a chain.  The data in the nodes is "forgotten".
 *
 * Returns:     []
 */
void chain_destroy (chain_t *chain);


/* -----------------------------------------------------------------
 * Function:    int chain_length (const chain_t *chain)
 *
 * Description: Get the number of nodes in chain
 *
 * Returns:     The number of nodes in chain
 */
int chain_length (const chain_t *chain);


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_first (const chain_t *chain)
 *
 * Description: Get the first node of chain
 *
 * Returns:     Pointer to first node in chain
 */
void *chain_first_all (const chain_t *chain, int value_flag);
#define chain_first(chain)		chain_first_all(chain, 0)
#define chain_first_data(chain)	chain_first_all(chain, 1)


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_last (const chain_t *chain)
 *
 * Description: Get the last node of chain
 *
 * Returns:     Pointer to last node in chain
 */
void *chain_last_all (const chain_t *chain, int value_flag);
#define chain_last(chain)		chain_last_all(chain, 0)
#define chain_last_data(chain)	chain_last_all(chain, 1)



/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_next (const chain_t *chain,
 *									      const chain_node_t *node)
 *
 * Description: Get the successor of a node
 *
 * Returns:     Pointer to next node in chain
 */
chain_node_t *chain_next (const chain_t *chain, const chain_node_t *node);


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_prev (const chain_t *chain,
 *									      const chain_node_t *node)
 *
 * Description: Get the predecessor of a node
 *
 * Returns:     Pointer to previous node in chain
 */
chain_node_t *chain_prev (const chain_t *chain, const chain_node_t *node);


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_enqeue (chain_t *chain, void *data)
 *
 * Description: Create a new node to keep 'data' and insert at in the 
 *						tail of the chain
 *
 * Returns:     Pointer to new node
 */
chain_node_t *chain_enqueue (chain_t *chain, void *data);


/* -----------------------------------------------------------------
 * Function:    void *chain_dequeue (chain_t *chain)
 *
 * Description: Remove the node at the heads of the chain
 *
 * Returns:     Pointer to node's data
 */
void *chain_dequeue (chain_t *chain);


/* -----------------------------------------------------------------
 * Function:    void *chain_dequeue_cond (chain_t *chain,
 *										  chain_walker_f *cond,
 *										  void *args)
 *
 * Description: Remove the first node from the chain for which the cond
 *				returns negative.
 *
 * Returns:     Pointer to node's data
 */
void *chain_dequeue_cond (chain_t *chain, chain_walker_f walker, void *args);


/* -----------------------------------------------------------------
 * Function:    void *chain_remove (chain_t *chain, chain_node_t *node)
 *
 * Description: Remove a node from the chain
 *
 * Returns:     Pointer to node's data
 */
void *chain_remove (chain_t *chain, chain_node_t *node);


/* -----------------------------------------------------------------
 * Function:    chain_t *chain_merge (chain_t *chain, chain_t *chain2)
 *
 * Description: Append a copy of all the elements in 'chain2' to 'chain'
 *
 * Returns:     Pointer to 'chain'
 */
chain_t *chain_merge (chain_t *chain, chain_t *chain2);


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_find (chain_t *chain,
 *									      const void *what,
 *								          chain_cmp_f compare,
 *								          void *args)
 *
 * Description: Lookpuk for 'what' in chain, using 'compare()' to
 *						find equality.
 *				'compare()' should return the same as 'strcmp()'
 *
 * Returns:     Pointer to node with equal data, or NULL if not found.
 */
chain_node_t *chain_find (chain_t *chain, const void *what,
					chain_compare_f compare, void *args);


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_walk (chain_t *chain,
 *						 		    chain_walk_f walker,
 *								    void *args)
 *
 * Description: Traverses the chain from head to tail applying the
 *				'walker()' function to the data object in each node.
 *				Stops immediately if 'wlaker()' returns negative.
 *
 * Returns:     Pointer to the node where walking stopped, or NULL if
 *				all the chain was traversed.
 */
chain_node_t *chain_walk (chain_t *chain, chain_walker_f walker, void *args);


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_rwalk (chain_t *chain,
 *										   chain_walk_f walker,
 *										   void *args)
 *
 * Description: Traverses the chain from tail to head applying the
 *						'walker()' function to the data object in
 *						each node.
 *				Stops immediately if 'walker()' returns negative.
 *
 * Returns:     Pointer to the node where walking stopped, or NULL if
 *				all the chain was traversed.
 */
chain_node_t *chain_rwalk (chain_t *chain, chain_walker_f walker, void *args);


/* -----------------------------------------------------------------
 * Function:    chain_node_t *chain_prune (chain_t *chain,
 *										   chain_walk_f pruner,
 *										   void *args)
 *
 * Description: Remove all nodes from chain for which 'pruner()' 
 *						returns positive.
 *				Stops the pruning if 'pruner()' returns negative.
 *				If the data in the node was dynamically allocated, the
 *						'pruner()' can release that memory before returning.
 * Returns:     Pointer to the node where prunning stopped, or NULL if
 *				all the chain was traversed.
 */
chain_node_t *chain_prune (chain_t *chain, chain_walker_f pruner, void *args);


/* -----------------------------------------------------------------
 * Function:    chain_t *chain_walk_collect (chain_t *chain,
 *										     chain_walk_collect_f walker,
 *											 void *args)
 *
 * Description: Traverses the chain from head to tail applying the
 *				'walker()' function to the data object in each node.
 *				Collects the pointers returned by 'walker()' in a chain.
 *				Stops immediately if 'wlaker()' returns NULL.
 *
 * Returns:     List with pointers returned by 'walker()'..
 */
chain_t *chain_walk_collect (chain_t *chain,
							 chain_walker_collect_f walker, void *args);


/* -----------------------------------------------------------------
 * Function:    void *chain_collect_origin (const chain_collect_t *node)
 *
 * Description: Get the origin kept in a collect node
 *
 * Returns:     Pointer to the origin in the collect node
 */
void *chain_collect_origin (const chain_collect_t *node);


/* -----------------------------------------------------------------
 * Function:    void *chain_collect_data (const chain_collect_t *node)
 *
 * Description: Get the data kept in a collect node
 *
 * Returns:     Pointer to the data in the collect node
 */
void *chain_collect_data (const chain_collect_t *node);



#endif /* __CHAIN_H */
